/*
 * Copyright 1992 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Athena Color Editor (the color editor from hell..)
 *
 * 7-30-92 Dave Sternlicht --  Creation.
 * 8-25-92 Dave Sternlicht --  Adding Color client data struct.
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Shell.h>
#include <X11/Xcms.h>
#include "CMSObject.h"
#include "ace.icon"

#include <stdio.h>

#define NUM_DIMENSIONS_PER_COLOR_SPACE 3
#define NUM_COLOR_SPACES 8
#define NUM_COLOR_DIMENSIONS 24

/* types */

/* these are lower case to distinguish from the XCms struct fields */
typedef enum  {rgb, rgbi, xyz, uvy, xyy, lab, luv, tekhvc} ColorSpaceType;

typedef struct {
  int source_space;
  ColorSpaceType colorSpace;

  Widget colorpatch;
  Widget cmsobj;
  /*** new ***/
  XcmsColor *cmsptr;
  Widget *space_scrollers;
  Widget *space_values;
  unsigned short **short_dimensions;
  XcmsFloat **float_dimensions;
  /*** old ***/
  XcmsCIEXYZ *ciexyz;
  Widget *color_boxes; /* ptr to an array of NUM_COLOR_SPACES widgets*/ 
  Widget *scrollers;   /* ptr to an array of NUM_COLOR_DIMENSIONS widgets*/ 
  Widget *values;      /* ptr to an array of NUM_COLOR_DIMENSIONS widgets*/ 
  XcmsCCC defCCC;
} ColorSpaceData, *ColorSpaceDataPtr;

/* constants */

static char *color_names[] = {
  "RGB   ", "RGBi  ", "XYZ   ", "uvY   ", "xyY   ", "Lab   ", "Luv   ",
  "TekHVC"
};

static char dimension_names[] = { 'R', 'G', 'B', 'R', 'G', 'B', 'X', 'Y', 'Z',
 'u', 'v', 'Y', 'x', 'y', 'Y', 'L', 'a', 'b', 'L', 'u', 'v', 'H', 'V', 'C'};

static XcmsFloat color_bounds[2][NUM_COLOR_SPACES * 3] = {
  {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, -2, -2, 0, 0, 0},
  {65535, 65535, 65535, 1, 1, 1, 1, 1, 1.5, 1, 1, 1, 1, 1, 1, 1,  1,  1, 1, 2,
   2, 360, 100, 100}
};

static XcmsColorFormat color_formats[NUM_COLOR_SPACES] = { 
  XcmsRGBFormat,
  XcmsRGBiFormat,
  XcmsCIEXYZFormat,      
  XcmsCIEuvYFormat,	
  XcmsCIExyYFormat,	
  XcmsCIELabFormat,	
  XcmsCIELuvFormat,	
  XcmsTekHVCFormat	
};

/* forward declarations */

static void ConvertToXYZ(),     UpdateTextView(),    UpdateSliderView(), 
            CMSObserverCB(),    ToggleCB(),          ScrollerCB(), 
            InitializeValues(), InitializeWidgets();

static float GetColorVal();  

static Pixel ConvertToRGBPixel();

static unsigned int GetColorFormat();

/*
 * GetColorVal() -- Get a color value from a scrollers resources.
 */
static float GetColorVal(cptr, i)
  ColorSpaceDataPtr cptr;
  int i;
{
  Arg warg[1];
  float percent;
  Widget* scrollers = cptr->scrollers;
  
  XtSetArg(warg[0], XtNtopOfThumb, &percent);
  XtGetValues(scrollers[i], warg, 1);
  return (color_bounds[1][i] - color_bounds[0][i])*percent;
}

/*
 * ConvertToRGBPixel() -- Convert from an XCMS color space to good old RGB.
 */
static Pixel ConvertToRGBPixel(cptr, w, color)
  ColorSpaceDataPtr cptr;
  Widget w;
  XcmsColor color;
{
  Display* dpy = XtDisplay(w);
  int scr = DefaultScreen(dpy);  

  XColor foo[1];
  if (
    XcmsFailure == XcmsConvertColors(
      cptr->defCCC, &color, 1, XcmsRGBFormat, NULL
    )
  ) printf("ConvertToRGBPixel: color conversion failure\n"); 
  foo[0].red = color.spec.RGB.red;    
  foo[0].green = color.spec.RGB.green;    
  foo[0].blue = color.spec.RGB.blue;    
  foo[0].flags = DoRed | DoGreen | DoBlue;
  if (!XAllocColor(dpy, DefaultColormap(dpy, scr), foo))
    printf("ConvertToRGBPixel: XAllocColor failed\n");
  XFreeColors(dpy, DefaultColormap(dpy, scr), &foo[0].pixel, 1, 0);
  return foo[0].pixel;
}

/*
 * ConvertToXYZ() -- Convert an XCMS color struct to XYZ.
 */
static void ConvertToXYZ(cptr, color)
  ColorSpaceDataPtr cptr;      
  XcmsColor* color;
{
  if (
    XcmsFailure == XcmsConvertColors(
      cptr->defCCC, color, 1, XcmsCIEXYZFormat, NULL
    )
  ) printf("ConvertToXYZ: color conversion failure\n"); 
}

/*
 * ExitCB() -- Je suis fini.
 */
static void
ExitCB(w, clientData, callData) /*ARGSUSED*/
  Widget w; XtPointer clientData, callData;
{
  exit(0);
}

/*
 * GetColorFormat() -- determin the Xcms color space format.
 */
static unsigned int
GetColorFormat(dimensionSpace) 
     ColorSpaceType dimensionSpace;
{
  return color_formats[dimensionSpace];
}

/*
 * UpdateTextView() -- update text based on slider.
 */
#define SET_COLOR_TEXT(index, format, space, stim)       \
  sprintf(text_value, format, colorRec.spec.space.stim); \
  XtSetArg(warg[0], XtNstring, text_value);              \
  XtSetValues(values[index], warg, 1)

static void
UpdateTextView(cptr, colorSpace, colorRec)
    ColorSpaceDataPtr cptr;
    ColorSpaceType colorSpace;
    XcmsColor colorRec;
{
  Widget* values = cptr->values;
  Arg warg[1];
  char text_value[100];  
  switch (colorSpace) {
  case rgb:
    SET_COLOR_TEXT(0, "%x", RGB, red);
    SET_COLOR_TEXT(1, "%x", RGB, green);
    SET_COLOR_TEXT(2, "%x", RGB, blue);
    break;
  case rgbi:
    SET_COLOR_TEXT(3, "%f", RGBi, red);
    SET_COLOR_TEXT(4, "%f", RGBi, green);
    SET_COLOR_TEXT(5, "%f", RGBi, blue);
    break;
  case xyz:
    SET_COLOR_TEXT(6, "%f", CIEXYZ, X);
    SET_COLOR_TEXT(7, "%f", CIEXYZ, Y);
    SET_COLOR_TEXT(8, "%f", CIEXYZ, Z);
    break; 
  case uvy:
    SET_COLOR_TEXT(9, "%f", CIEuvY, u_prime);
    SET_COLOR_TEXT(10, "%f", CIEuvY, v_prime);
    SET_COLOR_TEXT(11, "%f", CIEuvY, Y);
    break; 
  case xyy:
    SET_COLOR_TEXT(12, "%f", CIExyY, x);
    SET_COLOR_TEXT(13, "%f", CIExyY, y);
    SET_COLOR_TEXT(14, "%f", CIExyY, Y);
    break; 
  case lab:
    SET_COLOR_TEXT(15, "%f", CIELab, L_star);
    SET_COLOR_TEXT(16, "%f", CIELab, a_star);
    SET_COLOR_TEXT(17, "%f", CIELab, b_star);
    break; 
  case luv:
    SET_COLOR_TEXT(18, "%f", CIELuv, L_star);
    SET_COLOR_TEXT(19, "%f", CIELuv, u_star);
    SET_COLOR_TEXT(20, "%f", CIELuv, v_star);
    break; 
  case tekhvc:
    SET_COLOR_TEXT(21, "%f", TekHVC, H);
    SET_COLOR_TEXT(22, "%f", TekHVC, V);
    SET_COLOR_TEXT(23, "%f", TekHVC, C);
    break; 
  }
}

/*
 * UpdateSliderView() -- update slider based on slider.
 */
#define SET_COLOR_SLIDER(index, space, stim)                                  \
  XawScrollbarSetThumb(                                                       \
    scrollers[index],                                                         \
    (XcmsFloat)colorRec.spec.space.stim / (XcmsFloat)color_bounds[1][index],  \
    -1.0                                                                      \
  )

static void
UpdateSliderView(cptr, colorSpace, colorRec) 
    ColorSpaceDataPtr cptr;
    ColorSpaceType colorSpace;
    XcmsColor colorRec;
{
  Widget* scrollers = cptr->scrollers;
  switch (colorSpace) {
  case rgb: 
    SET_COLOR_SLIDER(0, RGB, red);
    SET_COLOR_SLIDER(1, RGB, green);
    SET_COLOR_SLIDER(2, RGB, blue);
  break;
  case rgbi: 
    SET_COLOR_SLIDER(3, RGBi, red);
    SET_COLOR_SLIDER(4, RGBi, green);
    SET_COLOR_SLIDER(5, RGBi, blue);
  break;
  case xyz: 
    SET_COLOR_SLIDER(6, CIEXYZ, X);
    SET_COLOR_SLIDER(7, CIEXYZ, Y);
    SET_COLOR_SLIDER(8, CIEXYZ, Z);
  break;
  case uvy:
    SET_COLOR_SLIDER(9, CIEuvY, u_prime);
    SET_COLOR_SLIDER(10, CIEuvY, v_prime);
    SET_COLOR_SLIDER(11, CIEuvY, Y);
  break;
  case xyy:
    SET_COLOR_SLIDER(12, CIExyY, x);
    SET_COLOR_SLIDER(13, CIExyY, y);
    SET_COLOR_SLIDER(14, CIExyY, Y);
  break;
  case lab:
    SET_COLOR_SLIDER(15, CIELab, L_star);
    SET_COLOR_SLIDER(16, CIELab, a_star);
    SET_COLOR_SLIDER(17, CIELab, b_star);
  break;
  case luv:
    SET_COLOR_SLIDER(18, CIELuv, L_star);
    SET_COLOR_SLIDER(19, CIELuv, u_star);
    SET_COLOR_SLIDER(20, CIELuv, v_star);
  break;
  case tekhvc:
    SET_COLOR_SLIDER(21, TekHVC, H);
    SET_COLOR_SLIDER(22, TekHVC, V);
    SET_COLOR_SLIDER(23, TekHVC, C);
  }
}

/*
 * CMSObserverCB() -- frobs the scrollers and text areas of the given
 *                          color space.
 */
static void
CMSObserverCB(w, clientData, callData) /*ARGSUSED*/
     Widget w; XtPointer clientData, callData;
{
  ColorSpaceType colorSpace = (ColorSpaceType)clientData;
  ColorSpaceDataPtr cptr = (ColorSpaceDataPtr)callData;   
  XcmsCIEXYZ* ciexyz = cptr->ciexyz;
  XcmsColor colorIO;

  /* ok to skip if the colorSpace is the same as the source space */
  if (colorSpace == cptr->source_space) return; 
  colorIO.spec.CIEXYZ = *ciexyz;
  colorIO.format = XcmsCIEXYZFormat;

  /* convert from xyz to the appropriate colorSpace */
  /* convert the color NEED A GENERAL ROUTINE...*/
  if (
    XcmsFailure == XcmsConvertColors(
      cptr->defCCC, &colorIO, 1, color_formats[colorSpace], NULL
    )
  ) printf("CMSObserverCB: color conversion failure\n");  

  /* update the text */
  UpdateTextView(cptr, colorSpace, colorIO);
  /* update the scrollers */
  UpdateSliderView(cptr, colorSpace, colorIO);
}

/*
 * ToggleCB() -- add/delete an observer from the observer callback list.
 */
static void
ToggleCB(w, clientData, callData) /*ARGSUSED*/
     Widget w; XtPointer clientData, callData;
{
  ColorSpaceDataPtr cptr = (ColorSpaceDataPtr)clientData;   
  ColorSpaceType cspace = cptr->colorSpace;
/*  ColorSpaceType cspace = (ColorSpaceType)clientData; */
  if (!XtIsSensitive(cptr->color_boxes[cspace])) {  
    /* add an observer callback */
    XtSetSensitive(cptr->color_boxes[cspace], True);
    XtAddCallback(cptr->cmsobj, XtNcallback, CMSObserverCB, (XtPointer)cspace);
  }
  else { /* delete an observer callback */
    XtRemoveCallback(cptr->cmsobj, XtNcallback, 
      CMSObserverCB, (XtPointer)cspace);
    XtSetSensitive(cptr->color_boxes[cspace], False);
  }
}

/*
 * ScrollerCB() -- Setup an Xcms structure with the colorSpace values.
 *                 Update the visual color display. Convert to CIEXYZ.
 *                 Notify all other views to update because data has changed.
 */
static void
ScrollerCB(w, clientData, callData) /*ARGSUSED*/
     Widget w; XtPointer clientData, callData;
{
  ColorSpaceDataPtr cptr = (ColorSpaceDataPtr)clientData; 
  Arg warg[1];  
  Pixel pix;
  XcmsColor colorIO;
  ColorSpaceType cspace = cptr->colorSpace;

  cptr->source_space = cspace;

  colorIO.format = GetColorFormat(cspace); 
  
  switch (cspace) {
    case rgb: 
      colorIO.spec.RGB.red = (unsigned short)GetColorVal(cptr, 0); 
      colorIO.spec.RGB.green = (unsigned short)GetColorVal(cptr, 1);
      colorIO.spec.RGB.blue = (unsigned short)GetColorVal(cptr, 2);
      break;
    case rgbi:
      colorIO.spec.RGBi.red = GetColorVal(cptr, 3); 
      colorIO.spec.RGBi.green = GetColorVal(cptr, 4);
      colorIO.spec.RGBi.blue = GetColorVal(cptr, 5);
      break;
    case xyz:
      colorIO.spec.CIEXYZ.X = GetColorVal(cptr, 6);
      colorIO.spec.CIEXYZ.Y = GetColorVal(cptr, 7);
      colorIO.spec.CIEXYZ.Z = GetColorVal(cptr, 8);
      break;
    case uvy:
      colorIO.spec.CIEuvY.u_prime = GetColorVal(cptr, 9); 
      colorIO.spec.CIEuvY.v_prime = GetColorVal(cptr, 10);
      colorIO.spec.CIEuvY.Y = GetColorVal(cptr, 11);
      break;
    case xyy:
      colorIO.spec.CIExyY.x = GetColorVal(cptr, 12);
      colorIO.spec.CIExyY.y = GetColorVal(cptr, 13);
      colorIO.spec.CIExyY.Y = GetColorVal(cptr, 14);
      break;
    case lab:
      colorIO.spec.CIELab.L_star = GetColorVal(cptr, 15);
      colorIO.spec.CIELab.a_star = GetColorVal(cptr, 16);
      colorIO.spec.CIELab.b_star = GetColorVal(cptr, 17);
      break;
    case luv:
      colorIO.spec.CIELuv.L_star = GetColorVal(cptr, 18);
      colorIO.spec.CIELuv.u_star = GetColorVal(cptr, 19);
      colorIO.spec.CIELuv.v_star = GetColorVal(cptr, 20);
      break;
    case tekhvc:
      colorIO.spec.TekHVC.H = GetColorVal(cptr, 21); 
      colorIO.spec.TekHVC.V = GetColorVal(cptr, 22);
      colorIO.spec.TekHVC.C = GetColorVal(cptr, 23);
      break;
  }

  UpdateTextView(cptr, cspace, colorIO);

  pix = ConvertToRGBPixel(cptr, w, colorIO);
  XtSetArg(warg[0], XtNbackground, pix);
  XtSetValues(cptr->colorpatch, warg, 1);
  /* now update the CMSObj's CIEXYZ values */
  ConvertToXYZ(cptr, &colorIO);
  *cptr->ciexyz = colorIO.spec.CIEXYZ;
  /* setting the cmsobject resource which invokes its callback list . */
/*  XtSetArg(warg[0], XtNcmsColor, cptr->ciexyz); */
  XtSetArg(warg[0], XtNcmsColor, cptr); 
  XtSetValues(cptr->cmsobj, warg, 1);
}

/*
 * InitializeValues() -- set up the color value widgets.
 */
static void
InitializeValues(values) 
  Widget *values;
{
  Arg warg[1];
  int i;
  for (i=0; i < 3; i++) {
    XtSetArg(warg[0], XtNstring, "0");
    XtSetValues(values[i], warg, (Cardinal)1);
  }
  for (i=3; i < NUM_COLOR_DIMENSIONS; i++) {
    XtSetArg(warg[0], XtNstring, "0.000000");
    XtSetValues(values[i], warg, (Cardinal)1);
  }
}


#define FIRST_COLOR rgb
#define LAST_COLOR  tekhvc

/*
 * InitializeWidgets() -- set up the user interface.  Note that the
 *                        aggregate of widgets that make up a given
 *                        color space should probibly be encabsulated
 *                        as a widget.  Note that the sliders in the
 *                        various colorspaces should be initially set
 *                        through CIEXYZ because when there are stimulus
 *                        values with the same callibrated intensity
 *                        XCMS chooses the value in the range that has
 *                        the largest intensity.
 */
  
static void
InitializeWidgets(toplevel)
  Widget toplevel;
{
  char* namestr;
  ColorSpaceType cspace;
  int dim, index;
  Arg warg;
  Widget hpaned, 
         vpaned, 
         toggle_boxes[NUM_COLOR_SPACES],
         control_boxes[NUM_COLOR_DIMENSIONS],
         labels[NUM_COLOR_DIMENSIONS],
         toggles[NUM_COLOR_SPACES],
         cpatch,
         exit_btn,
         cmsObj;
  /*** old ***/
  Widget *colorBoxes = (Widget *)malloc(sizeof(Widget) * NUM_COLOR_SPACES);
  Widget *sCrollers = (Widget *)malloc(sizeof(Widget) * NUM_COLOR_DIMENSIONS);
  Widget *vAlues = (Widget *)malloc(sizeof(Widget) * NUM_COLOR_DIMENSIONS);

  XcmsCIEXYZ *cIEXYZPtr = (XcmsCIEXYZ*)malloc(sizeof(XcmsCIEXYZ));

  cmsObj = XtCreateWidget("cmsobj", cmsObjClass, toplevel, NULL, 0); 

  hpaned = XtCreateManagedWidget("hpaned", panedWidgetClass, toplevel, NULL,0);

  cpatch = 
    XtCreateManagedWidget("colorpatch", labelWidgetClass, hpaned, NULL, 0);

  vpaned = 
    XtCreateManagedWidget("vpaned", panedWidgetClass, hpaned, NULL, 0);

  for (cspace = FIRST_COLOR, index = 0; cspace <= LAST_COLOR; cspace++) {
    ColorSpaceDataPtr cptr;
    /* creating the data members of cmsObject */
    cptr = (ColorSpaceDataPtr)malloc(sizeof(ColorSpaceData));
    toggle_boxes[cspace] = 
      XtCreateManagedWidget("", boxWidgetClass, vpaned, NULL, 0);

    /* put a toggle in a box */
    namestr = (char *)malloc(strlen(color_names[cspace]) + 9);
    sprintf(namestr, "%s", color_names[cspace]);
    toggles[cspace] = 
      XtCreateManagedWidget(
        namestr,
        toggleWidgetClass,
        toggle_boxes[cspace],
        NULL, 0
      );

    XtAddCallback(toggles[cspace], XtNcallback, ToggleCB, (Arg*)cptr);

    XtFree(namestr);

    /* assign pointers to the widget arrays */
    cptr->color_boxes = colorBoxes;
    /*** old ***/
    cptr->scrollers = sCrollers;
    cptr->values = vAlues;
    /*** new ***/
    cptr->space_scrollers = (Widget *)malloc(sizeof(Widget) * 
      NUM_DIMENSIONS_PER_COLOR_SPACE);
    cptr->space_values = (Widget *)malloc(sizeof(Widget) * 
      NUM_DIMENSIONS_PER_COLOR_SPACE);



    
    /* save the default color convertion context */
    cptr->defCCC = XcmsDefaultCCC(
      XtDisplay(cpatch), DefaultScreen(XtDisplay(cpatch))
    );    

    cptr->color_boxes[cspace] =
      XtCreateManagedWidget("", boxWidgetClass, toggle_boxes[cspace], 
        NULL, 0);

    XtSetSensitive(cptr->color_boxes[cspace], False);

    for (dim = 1; dim <= 3; dim++) {

      control_boxes[cspace] = XtCreateManagedWidget(
        "", boxWidgetClass, cptr->color_boxes[cspace], NULL, 0
      );

      /* put a scroller in a box */
      namestr = (char *)malloc(strlen(color_names[cspace]) + 9);
      sprintf(namestr, "scroller%s%d", color_names[cspace], dim);
      cptr->scrollers[index] = 
        XtCreateManagedWidget(
          namestr,
          scrollbarWidgetClass,
          control_boxes[cspace],
          NULL, 0
        );

      cptr->colorSpace = cspace;
      cptr->ciexyz = cIEXYZPtr;
      cptr->colorpatch = cpatch;
      cptr->cmsobj = cmsObj;

      XtAddCallback(
        cptr->scrollers[index], XtNjumpProc, ScrollerCB, (Arg *)cptr
      );
      XtFree(namestr);

      /* put a label in a box */
      namestr = (char *)malloc(2);
      sprintf(namestr, "%c", dimension_names[index]);
      labels[index] =
        XtCreateManagedWidget(
          namestr,
          labelWidgetClass, 
          control_boxes[cspace],
          NULL, 0
        );
      XtFree(namestr);

      /* put a text widget in a box */
      namestr = (char *)malloc(strlen(color_names[cspace]) + 9);
      sprintf(namestr, "text%s%d", color_names[cspace], dim);      
      cptr->values[index] =
        XtCreateManagedWidget(
          namestr,
          asciiTextWidgetClass, 
          control_boxes[cspace],
          NULL, 0
        );
      XtFree(namestr);

      index++;

    }
  }
  /* add the exit button */
  exit_btn = XtCreateManagedWidget("Press here to Quit", commandWidgetClass,
    vpaned, NULL, 0);
  XtAddCallback(exit_btn, XtNcallback, ExitCB, NULL);
  InitializeValues(vAlues);
}

/*
 * Main program.
 */
void main(argc, argv)
     int argc;
     char **argv;
{
  XtAppContext app;
  Widget toplevel;
  Arg  wargs[1];
  Pixmap icon;

  toplevel = XtAppInitialize(&app, "Ace", NULL, 0,
			     &argc, argv, NULL, NULL, 0);

  InitializeWidgets(toplevel);
  XtRealizeWidget(toplevel);
  
  /* Tell wm about icon */
  icon = XCreateBitmapFromData( XtDisplay(toplevel),
	       XtWindow(toplevel), (char *)ace_bits, ace_width, ace_height);
  XtSetArg( wargs[0], XtNiconPixmap, icon);
  XtSetValues( toplevel, wargs, 1);

  XtAppMainLoop(app);
}
